#!/usr/bin/perl -w
#-*- mode: Cperl -*-
#
##############################################################################
#
# This program is needed if you  want to import ADIF QSOs into a 'qle'
# log. Logs are kept in a table, inside an SQLite3 database file.
#
# The program 'adifimport.pl' converts an ADIF dump to an SQLite dump.
# Next, to  import the  SQLite dump  into a 'qle'  log table,  use the
# program 'sqlite3'.
#
##############################################################################
#
# Written by: Jeremy Cowgar KB8LFA
# Tweaks: Wilbert Knol ZL2BSJ/PE7T
#
# Added error handling for no config file:                10/5/2006 ZL2BSJ
# Added conversion from <FREQ> in MHz to kHz:             11/8/2008   PE7T
#
#                                                         16/4/2009   PE7T
# Added a workaround for missing <EOH>
# Added a conversion from ADIF <BAND> in meters to frequency in kHz.
# Removed automatic use of qle.conf if no config file specified; user must
# specify a 'qle' config file.
# Changed the way command-line options are parsed.
# Updated built-in help
#
# Fixed bug in pattern match expression                   11/9/2009   PE7T
#
##############################################################################
#
# Licence:
#
# Copyright (C) 2005 Jeremy Cowgar KB8LFA and Wilbert Knol ZL2BSJ
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free  Software Foundation; either  version 2 of the  License, or
# (at your option) any later version.
#
# This program is distributed in the  hope that it will be useful, but
# WITHOUT  ANY   WARRANTY;  without  even  the   implied  warranty  of
# MERCHANTABILITY or  FITNESS FOR A  PARTICULAR PURPOSE.  See  the GNU
# General Public License for more details.
#
# You should  have received a copy  of the GNU  General Public License
# along  with  this  program;  if  not, write  to  the  Free  Software
# Foundation, Inc., 59 Temple  Place, Suite 330, Boston, MA 02111-1307
# USA
#
##############################################################################
#
#
#
use warnings;
use strict;
use English;

my $debug = 0;

my $configFile = "";
my $adifFile   = "";

my $count = 0;
my $a     = "";
foreach $a (@ARGV) {
    $debug = 1 if $a =~ /^(--debug|-d)/;
    show_help() if $a =~ /^(--help|-h)/;
    if ( $a =~ m/^(--config|-c)/ ) {
        $configFile = $ARGV[ $count + 1 ] 
	  if defined $ARGV[ $count + 1 ];
    }
    $count++;
}

# Program needs at least 3 parameters: '-c' 'config file' and
# 'adif file':
show_help_and_quit() if $count < 3;

# Last argument must be ADIF file name:
$adifFile = $ARGV[ $count - 1 ] 
  if defined $ARGV[ $count - 1 ];

die " Error: can\'t find config file \'$configFile\'\n"
  unless -R $configFile;
die " Error: can\'t find ADIF file \'$adifFile\'\n"
  unless -R $adifFile;

use Config::General;
my $conf = new Config::General(
    -ConfigFile      => $configFile,
    -InterPolateVars => 1,
) or die;

my %config = $conf->getall
  or die "Could not read config file";

####
# Read the ADIF file
open( ADIF, $adifFile )
  or die "Could not open ADIF file \'$adifFile\'\n";

# Eat up until the end of the header
#   TODO: make this search lower case
#while (<ADIF>) {
#    if (/EOH/) { last; }
#}

my $content     = "";
my $endOfHeader = 0;
my $qsosRead    = 0;

LINE: while (<ADIF>) {
    my $line = $_;

    if ( $endOfHeader != 1 ) {
      if ( $line =~ m/^<EOH/ ) {
	# Line starts with <EOH>...
	# We've found the QSO block.
	$endOfHeader = 1;
      }
      next LINE;
    }


    if ($endOfHeader) {
        $content .= $line;    # Store the QSO in string $content
        $qsosRead++;
    }
    $endOfHeader = 1 if $line =~ /^<EOH>/;
}
close(ADIF);

# Parse the ADIF file
my $pos = 0;
my @recs; 

my $eors=0;
while ( $pos != -1 ) {
    ( $pos, my $rec ) = parse_record( $pos, $content );
    if ( $rec == -1 ) { last; }
    push @recs, $rec;
    $eors++;
}

# Build SQL statements based on parsed ADIF data
print "Building SQL statement:\n" if $debug;
my $qsosWritten = 0;
my $warnings = 0;
foreach my $rec (@recs) {
    my $insert = "INSERT INTO $config{tableName}";
    my $fields = "";
    my $values = "";

    foreach ( @{ $config{adifMap} } ) {
        my ( $adifName, $fname, $opts ) = split( /:/, $_ );
        my $value=$rec->{$adifName};
	if (!(defined $value)) {
	  print "\nIn ADIF QSO:\n" if $debug;
#	  warn "\$value undefined for \$adifname=$adifName\n" if $debug;
	  foreach my $k ( keys(%{$rec}) ) {
	    print "$k=$rec->{$k} " if $debug;
	  }
	  if ($debug) {
	    print "\n[***WARNING: ADIF field \'$adifName\' ";
	    print "not found in this QSO. ";
	    print "Using empty value \'\'.]\n";
	  }
	  $value = "";
	  $warnings++;
	}

        if ( $adifName eq "qso_date" ) {
            if ( $opts eq "yyyy" ) {
                $value = substr $value, 0, 4;
            }
            elsif ( $opts eq "mm" ) {
                $value = substr $value, 4, 2;
            }
            elsif ( $opts eq "dd" ) {
                $value = substr $value, 6, 2;
            }
        }

        if ( $adifName eq "freq" ) {
            if ( $fname eq $config{isKhz} ) {
                if ( $value !~ /^(\d+)\.(\d+)$/ ) {
                    print "Error: value $value for \<FREQ\> must " . 
		      "be in MHz.\n";
                    print "       E.g: \'14.123450\'\n";
                    print "Can't convert $value to kHz.\n";
#                    die;
                }
                die "Could not extract MHz portion from $value\n"
                  unless defined $1;
                die "Could not extract kHz portion from $value\n"
                  unless defined $2;
                my $mhz = $1;
                my $khz = $2;
                $value = "$1" . substr $khz, 0, 3;
            }
        }

        if ( $adifName eq "band" ) {
            if ( $fname eq $config{isKhz} ) {

                # Convert  band in  meters  to frequency  in kHz:  The
                # conversion is according to the array 'bandLimits' in
                # the qle  config file.   This script picks  the lower
                # band limit for the frequency.
                my $v      = "";
                my $meters = "";
                foreach ( @{ $config{bandLimits} } ) {
                    ( $meters, my $fMin, my $fMax ) = split( /:/, $_ );
                    die "Could not extract band from QSO: $rec\n"
                      unless defined $value;
                    if ( $value =~ /^($meters)$/ ) { $v = $fMin; }
                }
                die "Can't find band \'$value\' in array 'bandLimits'\n"
                  if $v eq "";
                $value = $v;
            }
        }

        if ( $adifName eq "time_on" ) {
            if ( $opts =~ /(\d{1})/ ) {
                $value = substr $value, 0, $1;
            }
        }

        $fields .= "$fname, ";
        $values .= "'$value', ";
      }

    # TODO:  not  very  safe  doing  unchecked  substr  bounds  Remove
    # trailing ', ' from $fields and $values:
    $fields = substr $fields, 0, ( length $fields ) - 2;
    $values = substr $values, 0, ( length $values ) - 2;
    my $sql = "$insert ($fields) VALUES ($values)";
    print $sql, ";\n";
    $qsosWritten++;
  }

if ($debug) {
    print "\n\n";
    print "       SUMMARY:\n\n";
    print "ADIF log file used:     \'$adifFile\'\n";
    print "\'qle'\ config file used: \'$configFile\'\n";
    print "\n";
    print "QSOs read:        $qsosRead\n";
    print "<EOR> tags found: $eors\n";
    print "QSOs written:     $qsosWritten\n";
    print "Warnings:         $warnings\n";
    print "\n";
  }

foreach ( @{ $config{adifMap} } ) {
  my $found=0;
  my ( $adifName, $fname, $opts ) = split( /:/, $_ );
  foreach ( @{ $config{fieldNames} } ) {
    my $fieldName = $_;
    $found=1 if "$fieldName" eq "$fname";
  }
  if ($found==0 && $debug) {
    print "***WARNING: In the config file: \'adifMap=$adifName:$fname\'\n";
    print "            but \'$fname\' does not exist as a \'qle\' field\n";
    print "            according to array \'fieldNames\'.\n";
    print "            Importing to non-existent fields is likely to fail.\n";
    print "\n";
  }
}

if ($debug) {
    print "***Warning: Embedded debug info will cause errors when reading\n";
    print "            and executing the SQL dump from within \'sqlite3\'\n";
    print "            Use the \'--debug\' or \'-d\' options only for";
    print " testing.\n\n";
}
exit;
#
#
#
##############################################################################
#
# end of main
#
##############################################################################

##############################################################################
#
# parse_record
#
# Purpose:
#
#           Looks  at a  QSO  in an  ADIF  log. Splits  that QSO  into
#           components.   Stores the  components  into an  associative
#           array  as key/value  pairs. The  keys are  ADIF  tags. The
#           values are field content.
#
#           Example: '<BAND:3>20m' is stored as 'band'->'20m'.
#
#           (ADIF tag  names are converted to lower  case, then stored
#           as keys)
#
#           The log is passed to  parse_record() as a single string in
#           which  each QSO  is  a  sub-string that  ends  in <EOR>  A
#           supplied pointer indicates where in  the log the QSO to be
#           processed starts.
#
# Args:
#
#           1. Starting   position  of  the  QSO   to  be   processed.
#           2. String  containing the ADIF  log as a  concatenation of
#              QSOs.
#
# Returns:
#
#           1.   Position  of  last   character  of  the  QSO  in  the
#           log. Normally this  is the position of the  closing '>' of
#           the  <EOR> tag.
#
#           2.  Reference to  associative array of ADIF-tag/ADIF-value
#           pairs.  This array contains only the QSO just processed.
#
##############################################################################
#
#
#
sub parse_record {
    my $cPos    = shift;
    my $content = shift;

    # my %rec = {};
    my %rec;

    my $bPos = 0; # beginning of tag
    my $ePos = 0; # end position of tag

    while (( $bPos = index $content, "<", $cPos ) != -1
	   && ( $ePos = index $content, ">", $bPos ) != -1 )
      {
        my ( $tname, $csize ) =
          split( /:/, substr $content, $bPos + 1, $ePos - $bPos - 1 );
        my $val = "";
        $val  = substr $content, $ePos + 1, $csize if defined $csize;
        $cPos = $ePos;
        $cPos = $ePos + $csize if defined $csize;

	# Quit loop on <EOR>:
        if ( lc $tname eq "eor" ) { last; }
	# ..else store tag/value pair:
        $rec{ lc $tname } = $val;
      }

    if ( $bPos == -1 || $ePos == -1 ) {
      return ( -1, -1 );
    }
    return ( $cPos, \%rec );
}

##############################################################################
#
# show_help_and_quit
#
# Purpose:
#
#           Prints helpful  text to STDOUT  on how to use  the script,
#           then exits  execution. The subroutine  is typically called
#           when wrong or  insufficient parameters have been supplied,
#           e.g. no 'qle' config file or no ADIF log file.
#
# Args:     none.
#
# Returns:  void
#
#
##############################################################################
#
#
#
sub show_help_and_quit {
    print "\n";
    print " adifimport.pl -  Converts an ADIF log to SQL statements.\n\n";
    print " The SQL statements can be fed to the program 'sqlite3'. Doing\n";
    print " this will add the QSOs to an existing database log.\n\n";
    print " The script adifimport.pl attempts to open a qle config file,\n";
    print " looks up tableName and conversion instructions (adifMap), and\n";
    print " dumps QSOs as INSERT statements to your screen on STDOUT.\n\n";
    print " Help:\n";
    print "      ./adifimport.pl   -h\n";
    print "                       --help\n";
    print " Usage:\n";
    print "      ./adifimport.pl     -c      <config-file> <adif-file>\n";
    print "                         --config <config-file> <adif-file>\n";
    print "\n";
    print " To produce debugging output  -recommended for initial testing-\n";
    print " and a summary, use the \'--debug\' or the \'-d\' switch:\n";
    print "\n";
    print "     ./adifimport.pl  -d      -c      <config-file> <adif-file>\n";
    print "     ./adifimport.pl --debug --config <config-file> <adif-file>\n";
    print "\n";
    print " To capture the SQL output, use redirection. Example:\n\n";
    print "     ./adifimport.pl  -c <config-file> <adif-file> > <sql.out>\n";
    print "\n";
    print " Next, run \'sqlite3 <database-file.db>\' to open the database\n";
    print " and then use the command \'.read <sql.out>' to import the\n";
    print " new QSOs. Do not use SQL output that contains embedded debug\n";
    print " info. Always make a backup of your database file first.\n";
    print "\n";

    exit;
}
#
#
##############################################################################
